/************************************************************
Copyright (c) 1993 by Silicon Graphics Computer Systems, Inc.

Permission to use, copy, modify, and distribute this
software and its documentation for any purpose and without
fee is hereby granted, provided that the above copyright
notice appear in all copies and that both that copyright
notice and this permission notice appear in supporting
documentation, and that the name of Silicon Graphics not be
used in advertising or publicity pertaining to distribution
of the software without specific prior written permission.
Silicon Graphics makes no representation about the suitability
of this software for any purpose. It is provided "as is"
without any express or implied warranty.

SILICON GRAPHICS DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL SILICON
GRAPHICS BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
THE USE OR PERFORMANCE OF THIS SOFTWARE.

********************************************************/

#define	NEED_MAP_READERS
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include "Xlibint.h"
#include <X11/extensions/XKBproto.h>
#include "XKBlibint.h"
#include <stdio.h>

/***====================================================================***/

XkbDescPtr
XkbGetKeyboardByName(Display *dpy,
                     unsigned deviceSpec,
                     XkbComponentNamesPtr names,
                     unsigned want,
                     unsigned need,
                     Bool load)
{
    register xkbGetKbdByNameReq *req;
    xkbGetKbdByNameReply rep;
    int len, extraLen = 0;
    char *str;
    XkbDescPtr xkb;
    int mapLen, codesLen, typesLen, compatLen;
    int symsLen, geomLen;
    XkbInfoPtr xkbi;

    fprintf (stderr, "test %p\n", dpy);
    if (dpy)
        fprintf (stderr, "test2 %0x %p\n", dpy->flags, dpy->xkb_info);
    if ((dpy == NULL) || (dpy->flags & XlibDisplayNoXkb) ||
        (!dpy->xkb_info && !XkbUseExtension(dpy, NULL, NULL))) {
    fprintf (stderr, "test3\n");
        return NULL;
    }

    xkbi = dpy->xkb_info;
    xkb = (XkbDescRec *) _XkbCalloc(1, sizeof(XkbDescRec));
    if (!xkb)
        return NULL;
    xkb->device_spec = deviceSpec;
    xkb->map = (XkbClientMapRec *) _XkbCalloc(1, sizeof(XkbClientMapRec));
    xkb->dpy = dpy;

    fprintf (stderr, "test4\n");
    LockDisplay(dpy);
    fprintf (stderr, "test5\n");
    GetReq(kbGetKbdByName, req);
    fprintf (stderr, "test6 %d %d %d %d\n", xkbi->codes->major_opcode, xkb->device_spec, want, need);
    req->reqType = xkbi->codes->major_opcode;
    req->xkbReqType = X_kbGetKbdByName;
    req->deviceSpec = xkb->device_spec;
    req->want = want;
    req->need = need;
    req->load = load;

    mapLen = codesLen = typesLen = compatLen = symsLen = geomLen = 0;
    if (names) {
        if (names->keymap)
            mapLen = (int) strlen(names->keymap);
        if (names->keycodes)
            codesLen = (int) strlen(names->keycodes);
        if (names->types)
            typesLen = (int) strlen(names->types);
        if (names->compat)
            compatLen = (int) strlen(names->compat);
        if (names->symbols)
            symsLen = (int) strlen(names->symbols);
        if (names->geometry)
            geomLen = (int) strlen(names->geometry);
        if (mapLen > 255)
            mapLen = 255;
        if (codesLen > 255)
            codesLen = 255;
        if (typesLen > 255)
            typesLen = 255;
        if (compatLen > 255)
            compatLen = 255;
        if (symsLen > 255)
            symsLen = 255;
        if (geomLen > 255)
            geomLen = 255;
    }
    else
        mapLen = codesLen = typesLen = compatLen = symsLen = geomLen = 0;

    len = mapLen + codesLen + typesLen + compatLen + symsLen + geomLen + 6;
    fprintf (stderr, "test7 %d %d %d %d %d %d %d\n", len, mapLen, codesLen, typesLen, compatLen, symsLen, geomLen);
    len = XkbPaddedSize(len);
    fprintf (stderr, "test8 %d %d\n", len, req->length);
    req->length += len / 4;
    fprintf (stderr, "test9 %d\n", req->length);
    BufAlloc(char *, str, len);

    *str++ = mapLen;
    if (mapLen > 0) {
        memcpy(str, names->keymap, (size_t) mapLen);
        str += mapLen;
    }
    *str++ = codesLen;
    if (codesLen > 0) {
        memcpy(str, names->keycodes, (size_t) codesLen);
        str += codesLen;
    }
    *str++ = typesLen;
    if (typesLen > 0) {
        memcpy(str, names->types, (size_t) typesLen);
        str += typesLen;
    }
    *str++ = compatLen;
    if (compatLen > 0) {
        memcpy(str, names->compat, (size_t) compatLen);
        str += compatLen;
    }
    *str++ = symsLen;
    if (symsLen > 0) {
        memcpy(str, names->symbols, (size_t) symsLen);
        str += symsLen;
    }
    *str++ = geomLen;
    if (geomLen > 0) {
        memcpy(str, names->geometry, (size_t) geomLen);
        str += geomLen;
    }
    fprintf (stderr, "test10\n");
    if ((!_XReply(dpy, (xReply *) &rep, 0, xFalse)) || (!rep.reported)) {
    fprintf (stderr, "test11\n");
        goto BAILOUT;
    }
    extraLen = (int) rep.length * 4;

    fprintf (stderr, "test12 %x %d %d %d\n", rep.reported, rep.deviceID, rep.minKeyCode, rep.maxKeyCode);
    xkb->device_spec = rep.deviceID;
    xkb->min_key_code = rep.minKeyCode;
    xkb->max_key_code = rep.maxKeyCode;
    if (rep.reported & (XkbGBN_SymbolsMask | XkbGBN_TypesMask)) {
        xkbGetMapReply mrep;
        Status status;
        int nread = 0;

    fprintf (stderr, "test13\n");
        _XRead(dpy, (char *) &mrep, SIZEOF(xkbGetMapReply));
        extraLen -= SIZEOF(xkbGetMapReply);
    fprintf (stderr, "test14\n");
        status = _XkbReadGetMapReply(dpy, &mrep, xkb, &nread);
    fprintf (stderr, "test15 %d\n", status);
        extraLen -= nread;
        if (status != Success)
            goto BAILOUT;
    }
    if (rep.reported & XkbGBN_CompatMapMask) {
        xkbGetCompatMapReply crep;
        Status status;
        int nread = 0;

    fprintf (stderr, "test16\n");
        _XRead(dpy, (char *) &crep, SIZEOF(xkbGetCompatMapReply));
        extraLen -= SIZEOF(xkbGetCompatMapReply);
    fprintf (stderr, "test17\n");
        status = _XkbReadGetCompatMapReply(dpy, &crep, xkb, &nread);
    fprintf (stderr, "test18 %d\n", status);
        extraLen -= nread;
        if (status != Success)
            goto BAILOUT;
    }
    if (rep.reported & XkbGBN_IndicatorMapMask) {
        xkbGetIndicatorMapReply irep;
        Status status;
        int nread = 0;

    fprintf (stderr, "test19\n");
        _XRead(dpy, (char *) &irep, SIZEOF(xkbGetIndicatorMapReply));
        extraLen -= SIZEOF(xkbGetIndicatorMapReply);
    fprintf (stderr, "test20\n");
        status = _XkbReadGetIndicatorMapReply(dpy, &irep, xkb, &nread);
    fprintf (stderr, "test21 %d\n", status);
        extraLen -= nread;
        if (status != Success)
            goto BAILOUT;
    }
    if (rep.reported & (XkbGBN_KeyNamesMask | XkbGBN_OtherNamesMask)) {
        xkbGetNamesReply nrep;
        Status status;
        int nread = 0;

    fprintf (stderr, "test22\n");
        _XRead(dpy, (char *) &nrep, SIZEOF(xkbGetNamesReply));
        extraLen -= SIZEOF(xkbGetNamesReply);
    fprintf (stderr, "test23\n");
        status = _XkbReadGetNamesReply(dpy, &nrep, xkb, &nread);
    fprintf (stderr, "test24 %d\n", status);
        extraLen -= nread;
        if (status != Success)
            goto BAILOUT;
    }
    if (rep.reported & XkbGBN_GeometryMask) {
        xkbGetGeometryReply grep;
        Status status;
        int nread = 0;

    fprintf (stderr, "test25\n");
        _XRead(dpy, (char *) &grep, SIZEOF(xkbGetGeometryReply));
        extraLen -= SIZEOF(xkbGetGeometryReply);
    fprintf (stderr, "test26\n");
        status = _XkbReadGetGeometryReply(dpy, &grep, xkb, &nread);
    fprintf (stderr, "test27 %d\n", status);
        extraLen -= nread;
        if (status != Success)
            goto BAILOUT;
    }
    fprintf (stderr, "test28 %d\n", extraLen);
    if (extraLen > 0)
        goto BAILOUT;
    UnlockDisplay(dpy);
    fprintf (stderr, "test29\n");
    SyncHandle();
    fprintf (stderr, "test30\n");
    return xkb;
 BAILOUT:
    if (xkb != NULL)
        XkbFreeKeyboard(xkb, XkbAllComponentsMask, xTrue);
    fprintf (stderr, "test31 %d\n", extraLen);
    if (extraLen > 0)
        _XEatData(dpy, extraLen);
    fprintf (stderr, "test32\n");
    UnlockDisplay(dpy);
    fprintf (stderr, "test33\n");
    SyncHandle();
    return NULL;
}

XkbDescPtr
XkbGetKeyboard(Display *dpy, unsigned which, unsigned deviceSpec)
{
    return XkbGetKeyboardByName(dpy, deviceSpec, NULL, which, which, False);
}
